/*
 * $Id: umapic.c,v 1.1.1.1 2002/08/16 17:29:27 root Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ptrace.h>
#include <sys/wait.h>
#include <signal.h>

#include <sched.h>

#include "external_constants.h"
#include "lotol_fdesc.h"
#include "umapic.h"

#define DEBUG 

#include "lilo.h"

#define PAGE_SIZE	4096

#define MAX_IRQS 256	/* XXX */
#define MAX_PHYS_CPUS	16
#define MAX_LOGIC_CPUS	32


#ifdef SMP
static int pending_irqs[MAX_IRQS][MAX_CPU] = { { 0, 0, /*...*/ } , 
						{ 0, 0, /*...*/ }, /* ..*/ 
						{ 0, 0, /*...*/ } };

static int apic_pid;
static struct local_apic *apic[MAX_CPU + 1];	/* FIXME */


/* IPIs with shorthand */
int
reg_ICR_shorthand(unsigned long reg, unsigned long val, int offset)
{
	int ret;
	int target;
	int mask = 1 << (val & 0x1F);

	dfprintf("reg_ICR_shorthand: 0x%lX triggered\n", val & 0xc0000);
	dfprintf("reg_ICR_shorthand: triggering vector 0x%lX\n", val & 0xFF);

	switch(val & 0xC0000) {
	case (APIC_DEST_SELF):
		apic[offset]->irr[(val & 0xFF) >> 5].bitfield |= mask;

		/* Remember the originator of this IPI */
		pending_irqs[val & 0xFF][offset] |= (1 << offset);
		dfprintf("reg_ICR_shorthand: set pending irq for irq 0x%lx "
			 "and cpu %d\n", val & 0xFF, offset);

		dfprintf("reg_ICR_shorthand: kill to %d\n", offset);
		ret = kill(pids[offset], SIGURG);
		assert(0 <= ret);
		break;

	case (APIC_DEST_ALLINC):
		for (target = 0; pids[target] != -1; target++) {
			apic[target]->irr[(val & 0xFF) >> 5].bitfield |= mask;

			/* Remember the originator of this IPI */
			pending_irqs[val & 0xFF][target] |= (1 << offset);
			dfprintf("reg_ICR_shorthand: set pending irq for irq "
				 "0x%lx and cpu %d\n", val & 0xFF, offset);

			dfprintf("reg_ICR_shorthand: kill to %d\n", target);
			ret = kill(pids[target], SIGURG);
			assert(0 <= ret);
		}
		break;

	case (APIC_DEST_ALLBUT):
		for (target = 0; pids[target] != -1; target++) {
			if (target == offset)
				continue;

			apic[target]->irr[(val & 0xFF) >> 5].bitfield |= mask;

			/* Remember the originator of this IPI */
			pending_irqs[val & 0xFF][target] |= (1 << offset);
			dfprintf("reg_ICR_shorthand: set pending irq for irq "
				 "0x%lx and cpu %d for target %d\n", val & 0xFF,
				 offset, target);

			dfprintf("reg_ICR_shorthand: kill to %d\n", target);
			ret = kill(pids[target], SIGURG);
			assert(0 <= ret);
		}
		break;

	default:
		assert(0);
		/*NOTREACHED*/
	}

	return 1;
}

int
fixed_ipi(unsigned long reg, unsigned long val, int offset)
{
	int target;
	int i;
	int ret;

	target = apic[offset]->icr2.destination;

	/* Physical or logical destination mode? */
	if (apic[offset]->icr1.destination_mode) {

		/* clustered or flat? */
		if(apic[offset]->dfr.model & 0xF) {
			/* flat */
			i = 0;	
			while (target >>= 1) { 
				i++;
			}

			assert(0 <= i && i <= 7);
			target = i;
		} else {	
			/* clustered mode current not supported - cry! */
			fprintf(stderr, "fixed_ipi():  clustered APIC mode "
					"not supported!\n");
			assert(0);
		}
	} else
		assert(target < MAX_PHYS_CPUS); /* 16 cpus in phys. dst. mode */

	dfprintf("fixed_ipi: DM_FIXED triggering vector 0x%lx (%d/%d)\n",
			val & 0xFF, pids[target], target);

	/* Set apropriate bit in IRR */
	apic[target]->irr[(val & 0xFF) >> 5].bitfield |= (1 << (val & 0x1F));

	/* Remember the originator of this IPI */
	pending_irqs[val & 0xFF][target] |= (1 << offset);
	dfprintf("fixed_ipi: set pending irq for irq 0x%lx and cpu %d\n", val & 0xFF, offset);

	/* And signal IPI! */
	dfprintf("fixed_ipi: kill to %d\n", offset);
	ret = kill(pids[target], SIGURG);
	assert(0 <= ret);

	return 1;
}

int
init_ipi(unsigned long reg, unsigned long val, int offset)
{
	int ret;
	int target;

	/* no shorthand */
	assert((val & 0x0C0000) == 0);

	dfprintf("init_ipi: DM_INIT triggered\n");
	target = apic[offset]->icr2.destination;
	assert(0 < target && target < MAX_CPU); /* No INIT to BSP! */

	if (val & APIC_INT_ASSERT) {
		ret = kill(pids[target], SIGURG);
		assert(0 <= ret);
	} else {
		dfprintf("init_ipi: deasserting DM_INIT\n");
	}

	/* XXX Remember the originator of this IPI */
	apic[offset]->icr1.delivery_status = 0;

	return 1;
}

/* IPIs without shorthand */
int
reg_ICR_noshorthand(unsigned long reg, unsigned long val, int offset)
{
	int ret;

	switch (val & 0x0700) {
	case (APIC_DM_FIXED):
		dfprintf("reg_ICR: APIC_DM_FIXED\n");
		ret = fixed_ipi(reg, val, offset);
		assert(0 <= ret);
		break;

	case (APIC_DM_LOWEST):
		dfprintf("reg_ICR: APIC_DM_LOWEST not implemented yet!\n");
		break;

	case (APIC_DM_SMI):
		dfprintf("reg_ICR: APIC_DM_SMI not implemented yet!\n");
		break;

	case (APIC_DM_NMI):
		dfprintf("reg_ICR: APIC_DM_NMI not implemented yet!\n");
		break;

	case (APIC_DM_REMRD):
		dfprintf("reg_ICR: APIC_DM_REMRD not implemented yet!\n");
		break;

	case (APIC_DM_INIT):
		dfprintf("reg_ICR: APIC_DM_INIT\n");
		ret = init_ipi(reg, val, offset);
		assert(0 <= ret);
		break;

	case (APIC_DM_STARTUP):
		dfprintf("reg_ICR: APIC_DM_STARTUP not implemented yet!\n");
		break;

	case (APIC_DM_EXTINT):
		dfprintf("reg_ICR: APIC_DM_EXTINT not implemented yet!\n");
		break;

	default: 
		/* something's wrong, bail out! */
		fprintf(stderr, "reg_ICR(): illegal Delivery Mode 0x%lx\n",val);
		assert(0);
		/*NOTREACHED*/
	}

	return 1;
}

/* Dispatch writes to the ICR -- Interrupt Command Register */
int
reg_ICR(unsigned long reg, unsigned long val, int offset)
{
	int ret;

	/* The delivery status is directly set to 1 (busy) by the write stub */
	assert(apic[offset]->icr1.delivery_status == 1);

	/* Shorthand */
	if ((val & 0xC0000) != 0) {
		ret = reg_ICR_shorthand(reg, val, offset);
		assert(0 <= ret);
	} else {
		ret = reg_ICR_noshorthand(reg, val, offset);
		assert(0 <= ret);
	}

	return 1;
}

int
reg_EOI(unsigned long reg, unsigned long val, int offset)
{
	int i, j, irq;
	int mask;

	/*
	 * XXX HSHOEXER:  Such loops suck!  Fix this make this cleaner!
	 */
	/* Clear highest ISR bit */
	for (i = 7; 0 <= i; i--) {
		mask = 0x80000000;

		for (j = 32; 0 < j; j--) {
			if (apic[offset]->isr[i].bitfield & mask) {
				apic[offset]->isr[i].bitfield &= ~mask;
				break;
			} else {
				/* LINTs and timer */
				if (apic[offset]->lvt_timer.vector 
						== i * 32 + j) {
					apic[offset]->lvt_timer.delivery_status = 0;
					return 1;	
				}
			}

			mask >>= 1;
		}
		if (0 < j)
			break;
	}

	/* clear Send Pending! */
	irq = i * 32 + j - 1;	/* XXX HSHOEXER */
	for (i = 0; i < 32; i++) {
		mask = 1 << i;
		if (pending_irqs[irq][offset] & mask) {

			/* CPU i sent us (offset) this irq */
			pending_irqs[irq][offset] &= ~mask;

			/* Is this irq pending at some other CPU? */
			for (j = 0; j < MAX_CPU; j++) {
				if (j == offset) 
					continue;
				if (pending_irqs[irq][j] & mask)
					break;
			}

			/* If no, clear APIC_ICR_BUSY at CPU i */
			if (j == MAX_CPU) {
				apic[i]->icr1.delivery_status = 0;
				dfprintf("reg_EOI: clear APIC_ICR_BUSY for irq "
					 "0x%x at cpu %d\n", irq, i);
			} else {
				assert(j < MAX_CPU);
				dfprintf("reg_EOI: irq 0x%x from cpu %d is "
					 "pending somewhere else (at least at "
					 "cpu %d)\n", irq, i, j);
			}
		}
	}

	return 1;
}

/*
 * This is the local APIC emulation.  Some stuff -- discarding writes to
 * read-only register, etc. -- is directly done in the stub
 * (arch/i386-linux/hardware/umapic.c) but the actual emulation is done
 * here.
 */
int
do_APIC(int offset)
{
	int ret;
	char buffer[64];
	unsigned long reg;
	unsigned long val;

	assert(0 <= offset && offset < MAX_CPU);

	ret = read(FD_FROM_CPU0 + offset, &buffer, sizeof(buffer));
	assert(ret <= sizeof(buffer));

	sscanf(buffer, "%ld:%*d\n", &reg);
	sscanf(buffer, "%*d:%ld\n", &val);

	dfprintf("do_APIC: reg=0x%lx val=0x%lx offset=0x%x\n", reg, val,offset);

	switch (reg) {
	case APIC_EOI:
		ret = reg_EOI(reg, val, offset);
		assert(0 <= ret);
		break;

	case APIC_ICR:
		ret = reg_ICR(reg, val, offset);
		assert(0 <= ret);
		break;

	default:
		dfprintf("do_APIC: non-implemented register 0x%lx\n", reg);
		break;
	}

	return 0;
}

/*
 * Just wait for requests from CPUs.
 */
static void
umapic(int ncpu)
{
	fd_set rset;
	int i;
	int ret;


	dfprintf("APIC is up and emulating %d local APIC(s), pid=%d\n", 
			ncpu, getpid());

	while (1) { 
		FD_ZERO(&rset);
		for (i = 0; i < ncpu; i++)
			FD_SET(FD_FROM_CPU0 + i, &rset);

		ret = select(FD_FROM_CPU0 + ncpu, &rset, NULL, NULL, NULL);
		assert(0 < ret);

		for (i = 0; i < ncpu; i++) {
			if (FD_ISSET(FD_FROM_CPU0 + i, &rset)) {
				ret = do_APIC(i);
				assert(0 <= ret);
			}
		}
	}
}
#endif	/* SMP */


void
umapic_init(int ncpu)
{
#ifdef SMP
	int i;
	int ret;
	int fd;
	int pfd[2];
	char map_path[256];
	int map_fd;
	static char zero[0x10000];

	/* create file for I/O-APIC memory */
	sprintf(map_path, "/tmp/pageioapic%08x", getpid());
	map_fd = open(map_path, O_RDWR|O_CREAT|O_EXCL, 0700);
	assert(0 <= map_fd && map_fd < FD_LOTOL_MIN);

	ret = unlink(map_path);
	assert(0 <= ret);

	ret = dup2(map_fd, FD_IOAPIC_MEMORY);
	assert(ret == FD_IOAPIC_MEMORY);

	ret = close(map_fd);
	assert(0 <= ret);

	ret = write(FD_IOAPIC_MEMORY, zero, PAGE_SIZE);
	assert(ret == PAGE_SIZE);

	/* create local APIC files */
	for (i = 0; i < opt_c; i++) {
		sprintf(map_path, "/tmp/pagelocalapic%d%08x", i, getpid());
		map_fd = open(map_path, O_RDWR|O_CREAT|O_EXCL, 0700);
		assert(0 <= map_fd && map_fd < FD_LOTOL_MIN);

		ret = unlink(map_path);
		assert(0 <= ret);

		ret = dup2(map_fd, FD_LAPIC0 + i);
		assert(ret == FD_LAPIC0 + i);

		ret = close(map_fd);
		assert(0 <= ret);

		ret = write(FD_LAPIC0 + i, zero, PAGE_SIZE);
		assert(ret == PAGE_SIZE);
	}

	/* create communication channels */
	for (i = 0; i < opt_c; i++) {
		ret = pipe(pfd);
		assert(0 <= ret);
		assert(0 <= pfd[0] && pfd[0] < FD_LOTOL_MIN);
		assert(0 <= pfd[1] && pfd[1] < FD_LOTOL_MIN);

		ret = dup2(pfd[0], FD_FROM_CPU0 + i);
		assert(ret == FD_FROM_CPU0 + i);

		ret = close(pfd[0]);
		assert(0 <= ret);

		ret = dup2(pfd[1], FD_TO_LAPIC0 + i);
		assert(ret == FD_TO_LAPIC0 + i);

		ret = close(pfd[1]);
		assert(0 <= ret);
	}

	/* initialize apics to a sane state */
	for (i = 0; i < ncpu; i++) {
		struct local_apic *a;

		fd = FD_LAPIC0 + i;

		/* no fixed mapping */
		a = (struct local_apic *) mmap((void *) 0, LOCAL_APIC_SIZE,
				PROT_READ | PROT_WRITE | PROT_EXEC,
				MAP_SHARED, fd, 0);
		assert(a != MAP_FAILED);

		/* Some default values */
		a->id.phys_apic_id = i;
		a->version.version = 0x1;
		a->version.max_lvt = 3;
		a->dfr.model = 0xF;
		a->svr.spurious_vector = 0xff;
		a->svr.apic_enabled = 1;
		a->svr.focus_cpu = 1;
		a->lvt_timer.mask = 1;
		a->lvt_pc.mask = 1;
		a->lvt_lint0.mask = 1;
		a->lvt_lint1.mask = 1;
		a->lvt_error.mask = 1;

		ret = munmap((char *) a, LOCAL_APIC_SIZE);
		assert(0 <= ret);
	}
}

void
umapic_fork(int ncpu)
{
	int i;
	int fd;

	/* create apic process */
	apic_pid = fork();
	assert(0 <= apic_pid);

	if (apic_pid == 0) {
		for (i = 0; i < ncpu; i++) {
			fd = FD_LAPIC0 + i;

			apic[i] = (struct local_apic *) mmap((void *) 0,
					LOCAL_APIC_SIZE,
					PROT_READ | PROT_WRITE | PROT_EXEC,
					MAP_SHARED, fd, 0);
			assert(0 <= (int)apic[i]);
		}
		apic[i] = NULL;	/* FIXME */

		umapic(ncpu);
		/*NOTREACHED*/
	}
#endif	/* SMP */
}

/*
 * CPUs (ie child processes) map their local APIC to a private address
 * range therefore this needs to be done by the children after forking.
 */
void
umapic_CPU_init(int cpu, int ncpu)
{
#ifdef SMP
	int i;
	int ret;

	assert(0 <= cpu && cpu < 32);

	ret = dup2(FD_LAPIC0 + cpu, FD_APIC_MEMORY);
	assert(ret == FD_APIC_MEMORY);

	ret = close(FD_LAPIC0 + cpu);
	assert(0 <= ret);

	/* Get rid of other childrens fd's */
	for (i = 0; i < ncpu; i++) {
		if (i != cpu) {
			ret = close(FD_LAPIC0 + i);
			assert(0 <= ret);
		}
	}
#endif	/* SMP */
}

void
umapic_shutdown(int ncpu)
{
#ifdef SMP
	int status;
	int ret;
	int i;

	ret = kill(apic_pid, SIGTERM);
	assert(0 <= ret);

	ret = waitpid(apic_pid, &status, 0);
	assert(ret == apic_pid);

	for (i = 0; i < ncpu; i++) {
		ret = close(FD_LAPIC0 + i);
		assert(ret <= 0);

		ret = close(FD_FROM_CPU0 + i);
		assert(ret <= 0);
	
		ret = close(FD_TO_LAPIC0 + i);
		assert(ret <= 0);
	}
#endif	/* SMP */
}
